#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_PCD8544.h>

// 27/10/2018: Camera mode trigger bug fixed

// Going to use two external interrupts:
// Pin 2: rotary encoder clock
// Pin 3: beam sensor
// Other IO pins:
// Display: 4,5,6,7,10
// Rotary encoder/switch: 8, 9
// Arm switch: 11
// Mode switch: Analog pin 0

// Rotary encoder inputs
const int CLKpin = 2; 
const int DTpin = 8;  
const int SWpin = 9;

// Other inputs/outputs
const int SPDpin = 3;                   // photodiode pin  
const int ARMpin = 11;
const int FTRIGpin = 12;                // flash trigger output
const int CTRIGpin = 13;                // camera trigger output
const int MODEpin = 14;                 // operating mode switch   *** check pin number and update ***

// Display: Software SPI
// pin 4 - Serial clock out (SCLK)
// pin 5 - Serial data out (DIN)
// pin 6 - Data/Command select (D/C)
// pin 7 - LCD chip select (CS)
// pin 10 - LCD reset (RST)
Adafruit_PCD8544 display = Adafruit_PCD8544(4, 5, 6, 7, 10);

// Rotary encoder value
int DelayVal = 10;

// Updated by the ISR (Interrupt Service Routine)
volatile int EncoderPosn = 10;
volatile boolean BeamBreak = false;
volatile boolean VBeamTest = false;
volatile long BeamBrkTime = 0;

boolean armed = false;
boolean armSWclosed = false;
boolean beam_detect = false;
boolean modeSWclosed = false;
boolean op_mode = false;                        // false: trigger camera, true: trigger camera then flash
boolean camTrigger = false;

long CurrTime;

// ------------------------------------------------------------------
// Encoder Interrupt
// ------------------------------------------------------------------
void isr ()  {
  static unsigned long lastInterruptTime = 0;
  // If interrupts come faster than 5ms, assume it's a bounce and ignore
  if (millis() - lastInterruptTime > 5) {
    if (digitalRead(DTpin) == LOW) {
      EncoderPosn-- ; // Could be -5 or -10
    }
    else {
      EncoderPosn++ ; // Could be +5 or +10
    }

    // Restrict value from 0 to +999
    EncoderPosn = min(999, max(0, EncoderPosn));

    // Keep track of when we were here last (no more than every 5ms)
    lastInterruptTime = millis();
  }
}

// ------------------------------------------------------------------
// Beam Break Interrupt 
// ------------------------------------------------------------------
void isr2 ()  {

  if (VBeamTest == true) {
    static unsigned long LastEdgeTime = 0;
    // If interrupts come faster than 5ms, assume it's a bounce and ignore
    if (millis() - LastEdgeTime > 5) {
      if (digitalRead(SPDpin) == LOW) {
        BeamBreak = true;
        BeamBrkTime = millis();                             // Time of the break - use for getting delay accurate
      }
      // Keep track of when we were here last (no more than every 5ms)
      LastEdgeTime = millis();
    }
  }
}

//******************************************************************************

void setup() {
  // Debug output on serial monitor
  Serial.begin(9600);

  // Rotary encoder pins are inputs
  pinMode(CLKpin, INPUT);
  pinMode(DTpin, INPUT);

  // Switch is floating so use the in-built PULLUP so we don't need a resistor
  pinMode(SWpin, INPUT_PULLUP);
  // Do same for the ARM and MODE switches
  pinMode(ARMpin, INPUT_PULLUP);
  pinMode(MODEpin, INPUT_PULLUP);
  pinMode(SPDpin, INPUT);
  
  // Trigger outputs
  pinMode(FTRIGpin, OUTPUT);
  pinMode(CTRIGpin, OUTPUT);


  // Attach the routine for rotary encoder interrupts
  attachInterrupt(digitalPinToInterrupt(CLKpin), isr, LOW);

  // Attach the routine for beam break interrupts
  attachInterrupt(digitalPinToInterrupt(SPDpin), isr2, FALLING);

  // Initialise the display
  display.begin(84, 48);
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(BLACK);
  display.setCursor(0,0);
  display.println("Mode: Camera");
  display.print("Delay: ");
  display.print(DelayVal);
  display.println("ms");
  display.println("Beam: No");
  display.println("Armed: 0");
  display.display();
  
  // Ready to go!
  Serial.println("Start");
}

// *****************************************************************************
void loop() {
  
  // do all the switch checking and display stuff when the system's not armed: no time to do it when the system's armed!
  if (!armed) {
    if (EncoderPosn != DelayVal) {                        // if rotary encoder position has changed then update
      // Keep track of this new value
      DelayVal = EncoderPosn;
      Serial.print("Delay: ");
      Serial.println(DelayVal);
    }
    if ((!digitalRead(SWpin))) {                          // rotary encoder's switch pressed? If yes reset the delay time
      EncoderPosn = 10;
    }
    if (digitalRead(SPDpin)) {
      beam_detect = true;
    }
    else {
      beam_detect = false;  
    }

    if (!digitalRead(MODEpin)) {                          // check the system mode switch
      delay(5);                                           // 5ms debounce
      modeSWclosed = true;
    }
    if (modeSWclosed) {
      if (digitalRead(MODEpin)) {
        modeSWclosed = false;
        op_mode = !op_mode;
        Serial.print("Mode: ");
        Serial.println(op_mode);
      }
    }
    if (!digitalRead(ARMpin)) {                           // check to see if system is armed
      delay(50);                                          // 50ms debounce
      armSWclosed = true;
    }
    if (armSWclosed) {
      if (digitalRead(ARMpin)) {
        armSWclosed = false;
        armed = !armed;
        if (armed) {
          BeamBreak = false;
          VBeamTest = true;
        }
        Serial.print("Armed: ");
        Serial.println(armed);
      }
    }
    // now update the display  
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(BLACK);
    display.setCursor(0,0);
    if (op_mode) {
      display.println("Mode:  Flash");
    }
    else {
      display.println("Mode:  Camera");
    }
    display.print("Delay: ");
    display.print(DelayVal);
    display.println("ms");
    display.print("Beam:  ");
    if (beam_detect == true) {
      display.println("Yes");
    }
    else {
      display.println("No");
    }
    display.print("Armed: ");
    display.println(armed);
    display.display();
  }
  // if the system's armed we go through the triggering cycle  

  while (armed) {
    if (op_mode) {                                          // system is in flash mode, so open camera shutter (bulb)
      if (!camTrigger) {                                    // first time through - trigger the camera
        camTrigger = true;
        Serial.print("Flash mode, Triggering camera at: ");
        Serial.println(millis());
        digitalWrite(CTRIGpin,HIGH);
      }
    }
    if (BeamBreak) {
      Serial.print("Beam break detected: ");
      Serial.print(millis());
      Serial.println("ms");
      BeamBreak = false;
      while (CurrTime < (BeamBrkTime + DelayVal)) {         // loop until delay length after beam is broken
        CurrTime = millis();
      }
      if (!camTrigger) {                                    // if system in camera mode, trigger the camera
        camTrigger = true;
        digitalWrite(CTRIGpin,HIGH);                        // trigger the camera
        digitalWrite(FTRIGpin,HIGH);                        // trigger the flash
        Serial.print("Camera mode, triggering camera and flash at: ");
        Serial.println(millis());
        delay(25);
        digitalWrite(FTRIGpin,LOW);                         // end of flash trigger
        delay(475); 
        digitalWrite(CTRIGpin,LOW);                         // delay, then end camera trigger
      }
      else {
        digitalWrite(FTRIGpin,HIGH);                        // trigger the flash
        Serial.print("Triggering flash at: ");
        Serial.println(millis());
        delay(25);
        digitalWrite(FTRIGpin,LOW);                         // end of flash trigger
        delay(25); 
        digitalWrite(CTRIGpin,LOW);                         // and close the camera shutter
      }
      camTrigger = false;
      Serial.println("trigger cycle completed");
      armed = false;                                        // disarm system
    }
    else {
      if (!digitalRead(ARMpin)) {                           // check to see if system is still armed
        delay(5);                                           // 5ms debounce
        armSWclosed = true;
      }
      if (armSWclosed) {
        if (digitalRead(ARMpin)) {
          armSWclosed = false;
          armed = !armed;
          Serial.print("Armed: ");
          Serial.println(armed);
          VBeamTest = !VBeamTest;
        }
      }
    }
  }
}
